
def scheduler_port() -> int:
    return 1384


def scheduler_message_port() -> int:
    return 1386


def ui_port() -> int:
    return 1385


def broadcast_port() -> int:
    return 34305


def worker_pool_port() -> int:
    return 1387


def worker_start_port() -> int:
    return 6969


def message_proxy_port() -> int:
    return 22182
